
// catch2 includes
#define CATCH_CONFIG_MAIN  // This tells Catch to provide a main() - only do this in one cpp file
#include "catch.hpp"
#include "test_reaction.hpp"
#include "test_voxel.hpp"
#include "test_simulator.hpp"
